
#ifndef CFSRWCD_H
#define CFSRWCD_H

#include "arpc.h"
#include "sfsclient.h"
#include "sfsmisc.h"
#include "dhash_prot.h"
#include "dhash.h"

#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "cfsa_prot.h"
#include "debug.h"

#include "cfsrw.h"
#include "attrcache.h"
#include "fcache.h"
#include "snapshot.h"

// file/attr cache entry is considered recent if it is 120 seconds or
// less old
#define CACHE_RECENT 120    

typedef callback<void, nfs_fh3*>::ref cbfh_t;

class cfsrw;
class cfsa_signhead_res;

class cfsrwcd
{
private:
  sfsprog &_prog;
  chordID _vid;
  writeverf3 _verf3;
  
  cfsrw_attr_cache _ac;
  lrucache<nfs_fh3, file_cache *> _fc;
  str _cdir;
  
  ref<view> _view;
  cfsrw _cfsrw;
  qhash<sfs_aid, ptr<log> > _logs;

  bool uid (nfscall *sbp, chordID &u);
  bool readonly (nfscall *sbp);
  ptr<log> mylog (nfscall *sbp);
  
  void dispatch_view (nfscall *, vervec start, cfsrw_status);
  void dispatch_cfsrw (nfscall *, cfscallp c);

  void flushfattr (unsigned int proc, void *arg, void *res);
  void getfattr (unsigned int proc, sfs_aid aid, void *arg, void *res);
  void error (nfscall *sbp, nfsstat3 err);
  void reply (nfscall *, void *, nfs_fh3);
  void reply (nfscall *, void *);

  struct snaprec {
    bool set;
    bool building;
    unsigned ts;
    chordID location;
    vervec version;
    ptr<snapshot> image;
  };

  qhash<chordID, ptr<snaprec> > _snapshots;

  str  fh2fn(nfs_fh3 fh);

  void read_from_cache (nfscall *, file_cache *);
  void read_from_cache_open (nfscall *, file_cache *, ptr<aiofh>, int);
  void read_from_cache_read (nfscall *, file_cache *, ptr<aiobuf>,
                             ssize_t, int);

  void write_to_cache (nfscall *, file_cache *);
  void write_to_cache_open (nfscall *, file_cache *, ptr<aiofh>, int);
  void write_to_cache_write (nfscall *, file_cache *, ptr<aiobuf>,
                             ssize_t, int);

  void truncate_cache (nfscall *, file_cache *, uint64);
  void truncate_cache_open (nfscall *, file_cache *, uint64, ptr<aiofh>, int);
  void truncate_cache_truncate (nfscall *, file_cache *, uint64, int);

  void run_rpcs (file_cache *);
  bool dont_run_rpc (nfscall *nc);
  
  bool fcache_dispatch (nfscall *);
  void fcache_check_reply (nfscall *sbp, void *res);

  void flush_cache (file_cache *e, nfscall *sbp);
  void flush_cache_with_v (file_cache *, nfscall *, vervec, cfsrw_status);
  void flush_done (file_cache *e, nfscall *sbp, bool ok);

  void fill_cache (file_cache *, callback<void,bool,bool>::ref, nfscall *);
  void fill_cache_with_v (file_cache *, callback<void,bool,bool>::ref,
                          nfscall *, vervec, cfsrw_status);
  void fill_done (file_cache *e, bool done, bool ok);

  void file_cache_gc_remove_cb (int);
  static void file_closed (int) {}

  void sign_log_head_cb (sfs_aid writer, cfsrw_sign_cb_t cb,
                         cfsa_signhead_res *res, clnt_stat err);
  void sign_log_head (sfs_aid writer, chordID head,
                      const str &data, cfsrw_sign_cb_t cb);
 
  void set_cfsrw_sattr (cfsrw_sattr &ca, sattr3 &sa);
  uint32 access_check (cfsrw_ftype t, uint32 access_req);

  void getattr_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void access_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void lookup_dotdot_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void lookup_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void write_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void write_found_file (nfscall *sbp, cfscallp c, cfsrw_status status);
  void read_reply (nfscall *sbp, cfscallp getattr_c,
                   cfscallp c, cfsrw_status status);
  void read_found_file (nfscall *sbp, cfscallp c, cfsrw_status status);
  void setattr_reply (nfscall *sbp, cfscallp getattr_c,
                      cfscallp c, cfsrw_status status);
  void setattr_found_obj (nfscall *sbp, cfscallp c, cfsrw_status status);
  void readlink_reply (nfscall *sbp, cfscallp getattr_c,
                       cfscallp c, cfsrw_status status);
  void readlink_found_file (nfscall *sbp, cfscallp c, cfsrw_status status);
  void readdir_reply (nfscall *sbp, cfscallp getattr_c,
                      cfscallp c, cfsrw_status status);
  void readdir_found_dir (nfscall *sbp, cfscallp c, cfsrw_status status);
  void remove_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void rmdir_readdir (nfscall *sbp, bigint object,
                      cfscallp c, cfsrw_status status);
  void remove_found_attr (nfscall *sbp, bigint object,
                          cfscallp c, cfsrw_status status);
  void remove_found_file (nfscall *sbp, cfscallp c, cfsrw_status status);
  void rename_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void rename_readdir (nfscall *sbp, cfscallp rename_c,
                       cfscallp c, cfsrw_status status);
  void rename_to_attr (nfscall *sbp, cfscallp rename_c,
                       cfscallp c, cfsrw_status status);
  void rename_to (nfscall *sbp, bigint from, cfsrw_ftype from_type,
                  cfscallp c, cfsrw_status status);
  void rename_from_attr (nfscall *sbp, bigint from,
                         cfscallp c, cfsrw_status status);
  void rename_from (nfscall *sbp, cfscallp c, cfsrw_status status);
  void create (nfscall *sbp, cfscallp c);
  void create_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void create_found_dir (nfscall *sbp, cfscallp c, cfsrw_status status);
  void link_found_obj (nfscall *sbp, cfscallp c, cfsrw_status status);
  void create_unchecked_reply (nfscall *sbp, bigint object,
                               cfscallp c, cfsrw_status s);
  void create_unchecked_lookup (nfscall *sbp, cfscallp c, cfsrw_status s);
  void excldirop_reply (nfscall *sbp, cfscallp c, cfsrw_status status);
  void excldirop_cancel_reply (nfscall *sbp, bigint fh, nfsstat3 r,
                               cfscallp c, cfsrw_status status);
  void excldirop_cancel (nfscall *sbp, bigint fh, nfsstat3 r, cfscallp c);
  bool load_snapshot_and_continue (nfscall *sbp, cfscallp c);
  void load_snapshot (nfscall *sbp, cfscallp c,
                      chordID loc, ptr<snapshot> s);
  void build_snapshot_with_v (ptr<log> l, vervec v, cfsrw_status status);
  void snapshot_created (ptr<log> l, chordID sID, ptr<snapshot> image);

  static qhash<sfs_aid, ref<aclnt> > cahash;
  static ptr<aclnt> get_agent (sfs_aid aid);

public:
  static aiod* a;

  void got_view (callback<void, cfsrwcd*, nfs_fh3*>::ref, cfsrw_status);
  void dispatch (ref<nfsserv>, nfscall*);
  void dispatch_nfs (nfscall*);
 
  cfsrwcd (sfsprog &p, chordID view,
           callback<void, cfsrwcd*, nfs_fh3*>::ref cb);
  
  cfscallp get_cfscall (nfscall *sbp, vervec start);
  void build_snapshot (ptr<log> l);

  void file_cache_insert (nfs_fh3 fh) {
    file_cache *f = New file_cache(fh);
    _fc.insert(fh,f);
  } 

  file_cache *file_cache_lookup (nfs_fh3 fh) {
    file_cache **e = _fc[fh];
    if (e) {
      assert(*e);
      return *e;
    }
    return 0;
  }

  void file_cache_gc_remove (file_cache *e);

  static unsigned blocksize;
};


inline bigint 
nfs2cfsfh (nfs_fh3 *nfh) 
{
  bigint n;
  mpz_set_rawmag_be (&n, nfh->data.base (), nfh->data.size ());
  return n;
}

inline void
cfsfh2nfs (bigint n, nfs_fh3 *nfh)
{
  str raw = n.getraw();
  nfh->data.setsize(raw.len());
  memcpy (nfh->data.base(), raw.cstr(), raw.len());
}

inline void
apply_sattr (fattr3& f, sattr3& s)
{
  bool ctime = false;
  if (s.size.set) {
    f.size = (uint64) *s.size.val;
    ctime = true;
  }
  if (s.mode.set) {
    f.mode = (uint32) *s.mode.val;
    ctime = true;
  }
  if (s.mtime.set == SET_TO_CLIENT_TIME) {
    f.mtime = (nfstime3) *s.mtime.time;
    ctime = true;
  }
  else if (s.mtime.set == SET_TO_SERVER_TIME) {
    f.mtime.seconds = tsnow.tv_sec;
    f.mtime.nseconds = tsnow.tv_nsec;
    ctime = true;
  }
  if (s.uid.set) {
    f.uid = (uint32) *s.uid.val;
    ctime = true;
  }
  if (s.gid.set) {
    f.gid = (uint32) *s.gid.val;
    ctime = true;
  }
  if (ctime) {
    f.ctime.seconds = tsnow.tv_sec;
    f.ctime.nseconds = tsnow.tv_nsec;
  }
}

inline nfsstat3
get_nfsstat3 (cfsrw_status s)
{
  switch (s) {
    case CFSRW_OK:
      return NFS3_OK;
    case CFSRWERR_IO:
      return NFS3ERR_IO;
    case CFSRWERR_NOENT:
      return NFS3ERR_NOENT;
    case CFSRWERR_EXISTS:
      return NFS3ERR_EXIST;
    case CFSRWERR_NOTDIR:
      return NFS3ERR_NOTDIR;
    case CFSRWERR_STALE:
      return NFS3ERR_STALE;
    case CFSRWERR_NOLINK:
      return NFS3ERR_STALE;
    case CFSRWERR_PATHTOOLONG:
      return NFS3ERR_NAMETOOLONG;
    case CFSRWERR_NOTSUPP:
      return NFS3ERR_NOTSUPP;
    case CFSRWERR_EXCLCONT:
      return NFS3ERR_EXIST;
    default:
      break;
  }
  return NFS3ERR_IO;
}

inline const strbuf &
strbuf_cat(const strbuf &b, nfs_fh3 fh)
{
  str s = armor32(fh.data.base(), fh.data.size());
  b << s;
  return b;
}

inline ptr<fattr3>
new_attr()
{
  ptr<fattr3> fa = New refcounted<fattr3>;
  fa->mode = fa->nlink = fa->uid = fa->gid 
           = fa->size = fa->fsid = fa->fileid
           = fa->rdev.major = fa->rdev.minor = 0;
  fa->atime.seconds = fa->atime.nseconds = 0;
  fa->mtime.seconds = fa->mtime.nseconds = 0;
  fa->ctime.seconds = fa->ctime.nseconds = 0;
  return fa;
}

inline void
fill_fattr (bigint object, fattr3& f)
{
  if (f.type == NF3DIR)
    f.size = 512;
  f.fsid = 0;
  f.used = f.size;
  f.fileid = cfsfh2fid (object);
}

inline ptr<fattr3>
get_fattr3 (bigint object, cfsrw_fattr *ca)
{
  ptr<fattr3> fa = new_attr ();
  fa->type =
    (ca->type == CFSRW_DIR) ? NF3DIR 
                            : ((ca->type == CFSRW_LNK) ? NF3LNK : NF3REG);
  fa->mode = ca->mode;
  fa->nlink = ca->nlink;
  fa->uid = ca->uid;
  fa->gid = ca->gid;
  fa->size = ca->size;
  fa->mtime.seconds = ca->mtime;
  fa->ctime.seconds = ca->ctime;
  fa->atime.seconds = ca->ctime;
  fill_fattr (object, *fa);
  return fa;
}

#endif

