
#ifndef CFSRW_PROTO_H
#define CFSRW_PROTO_H

#include "refcnt.h"
#include "nfs3_prot.h" // rpc_bytes, dirlist3
#include "vervec.h"

class log;
class snapshot;

enum cfsrw_status {
  CFSRW_OK             = 0,
  CFSRWERR_IO          = 1,
  CFSRWERR_NOENT       = 2,
  CFSRWERR_EXISTS      = 3,
  CFSRWERR_ABORTED     = 4,
  CFSRWERR_NOTDIR      = 5,
  CFSRWERR_STALE       = 6,
  CFSRWERR_NOLINK      = 40, // no link record for given file handle
  CFSRWERR_PATHTOOLONG = 41, // target of symlink too long
  CFSRWERR_EXCLCONT    = 42, // exclusive dirop contention
  CFSRWERR_NOVIEW      = 100,
  CFSRWERR_NORECORD    = 101,
  CFSRWERR_NOLOGHEAD   = 102,
  CFSRWERR_NOTLOGHEAD  = 103,
  CFSRWERR_CORRUPTLOG  = 104,
  CFSRWERR_CONFLICT    = 105,
  CFSRWERR_WAIT        = 200,
  CFSRWERR_NOTSUPP     = 999,
};

enum cfsrw_proc {
  CFSRWPROC_NULL      = 0,
  CFSRWPROC_GETATTR   = 1,
  CFSRWPROC_GETPARENT = 2,
  CFSRWPROC_LOOKUP    = 3,
  CFSRWPROC_READ      = 4,
  CFSRWPROC_READLINK  = 5,
  CFSRWPROC_READDIR   = 6,
  CFSRWPROC_SETATTR   = 7,
  CFSRWPROC_WRITE     = 8,
  CFSRWPROC_CREATE    = 9,
  CFSRWPROC_MKDIR     = 10,
  CFSRWPROC_SYMLINK   = 11,
  CFSRWPROC_RENAME    = 12,
  CFSRWPROC_UNLINK    = 13,
  CFSRWPROC_LINK      = 14,
  CFSRWPROC_EXCLDIROP = 15,
};

enum cfsrw_ftype {
  CFSRW_REG = 0,
  CFSRW_DIR = 1,
  CFSRW_LNK = 2,
};

struct cfsrw_fattr {
  cfsrw_ftype type;
  uint32 mode;
  uint32 nlink;
  uint32 uid;
  uint32 gid;
  uint64 size;
  uint32 mtime;
  uint32 ctime;
};

class cfsrw_sattr {
public:
  int valid_size;
  uint64 size;
  int valid_mode;
  uint32 mode;
  int valid_mtime;
  uint32 mtime;
  int valid_uid;
  uint32 uid;
  int valid_gid;
  uint32 gid;

  cfsrw_sattr () {
    valid_size = valid_mode = valid_mtime = valid_uid = valid_gid = 0;
  }
  void serialize (strbuf&) const;
  void xinxin (xin*);
};

// CFSRW_GETATTR

struct cfsrw_getattr_arg {
  bigint object;
};

struct cfsrw_getattr_res {
  cfsrw_fattr attributes;
};

// CFSRW_LOOKUP

struct cfsrw_lookup_arg {
  bigint dir;
  str name;
};

struct cfsrw_lookup_res {
  bigint object;
};

// CFSRW_GETPARENT

struct cfsrw_getparent_arg {
  bigint object;
};

struct cfsrw_getparent_res {
  bigint object;
};

// CFSRW_READLINK

struct cfsrw_readlink_arg {
  bigint file;
};

struct cfsrw_readlink_res {
  str path;
};

// CFSRW_READ

struct cfsrw_read_arg {
  bigint file;
  uint64 offset;
  uint32 count;
};

struct cfsrw_read_res {
  uint32 count;
  bool eof;
  rpc_bytes<RPC_INFINITY> data;
};

// CFSRW_READDIR

struct cfsrw_readdir_arg {
  bigint dir;
  uint64 cookie;
  bigint cookieverf;
  uint32 count;
};

struct cfsrw_readdir_res {
  bigint cookieverf;
  uint32 entries;
  dirlist3 reply;
};

// CFSRW_SETATTR

struct cfsrw_setattr_arg {
  bigint object;
  cfsrw_sattr new_attributes;
};

// CFSRW_CREATE
// CFSRW_MKDIR

struct cfsrw_create_arg {
  bigint dir;
  str name;
  cfsrw_sattr attributes;
};

struct cfsrw_create_res {
  bigint object;
};

// CFSRW_SYMLINK

struct cfsrw_symlink_arg {
  bigint dir;
  str name;
  str linkto;
  cfsrw_sattr attributes;
};

struct cfsrw_symlink_res {
  bigint object;
};

// CFSRW_RENAME

struct cfsrw_rename_arg {
  bigint object;
  cfsrw_ftype type;
  bigint from_dir;
  str from_name;
  bigint to_dir;
  str to_name;
  bool target_exists;
  bigint target;
};

// CFSRW_UNLINK

struct cfsrw_unlink_arg {
  bigint object;
  cfsrw_ftype type;
  bigint dir;
  str name;
};

// CFSRW_WRITE

struct cfsrw_write_arg {
  bigint file;
  uint64 offset;
  uint32 count;
  rpc_bytes<RPC_INFINITY> data;
};

// CFSRW_LINK

struct cfsrw_link_arg {
  bigint object;
  bigint dir;
  str name;
};

// CFSRW_EXCLDIROP

struct cfsrw_excldirop_arg {
  bigint dir;
  str name;
  bool cancel;
};


// call structure

struct cfsrw_nores_res {
};

class cfscall;

typedef callback<void, ptr<cfscall>, cfsrw_status>::ptr cfsrw_cb;

class cfscall : public virtual refcount {
  cfsrw_proc _proc;
  cfsrw_cb _cb;
  void *_arg;
  void *_res;
  void memop (bool free);

public:
  cfscall () : _proc (CFSRWPROC_NULL), _cb (0) { memop (false); }
  cfscall (cfsrw_proc p, cfsrw_cb cb) : _proc (p), _cb (cb) { memop (false); }
  ~cfscall () { memop (true); }

  ptr<log> l;
  vervec start;
  vervec stop;
  ptr<snapshot> s;

  cfsrw_proc proc () const { return _proc; }
  void *arg () const { return _arg; }
  void *res () const { return _res; }
  void reply (cfsrw_status status) { _cb (mkref(this), status); }
};
typedef ptr<cfscall> cfscallp;



#endif

