
#ifndef CFSRW_LOG_H
#define CFSRW_LOG_H

#include "record.h"

typedef callback<void, sfs_pubkey2, sfs_sig2, bool>::ref cfsrw_sign_cb_t;
typedef callback<void, const str&, cfsrw_sign_cb_t>::ptr cfsrw_sign_t;

class log : public virtual refcount 
{
private:
  chordID _head;
  bool _readonly;
  chordID _root;

  bool _refreshed;
  bool _loading;
  vec<cbb> _appends_cb;
  vec<recordp> _appends_e;
  chordID _snapshot;
  bool _has_snapshot;
  bool _snapshot_new;
  ptr<log_record> _head_record;

  int _stores;
  int _failed_stores;
  vec<cbb> _store_cbs;

  cfsrw_sign_t sign;

  void store_head_after_append(bool);
  void head_stored(ptr<vec<cbb> >, ptr<head_log_record>, bool);
  void do_append(recordp head);
  
  void store_record(chordID, const str&, cbb, recordp e);
  void store_record_cb(cbb, dhash_stat, chordID);
  void get_record_cb(cb_recordp, int, str);

  void append_log_cb1(recordp, cbb, recordp);
  void append_log_cb2(chordID, cbb, recordp, bool);

  void load_head(recordp);
  void load_head_cb(cb_recordp, recordp);
  void store_head(recordp, const str, cbb, sfs_pubkey2, sfs_sig2, bool);
  
  static void dhash2log(cb_recordp, chordID, ptr<dhash_block>);

public:
  log(chordID, cfsrw_sign_t s = 0);
  log(recordp e, cfsrw_sign_t s = 0);

  ptr<log_record> log_head () const { return _head_record; }

  chordID head () const { return _head; }
  chordID root () const { return _root; }

  bool readonly () const { return _readonly; }
  
  void append(const vervec &v, recordp, cbb);
  void add_snapshot(chordID s);
  bool has_snapshot() const { return _has_snapshot; }
  chordID snapshot() const { return _snapshot; }

  static void locate(chordID, bool keyhashed, cb_recordp);
  static void prev_record(recordp, cb_recordp);
};

#endif

