
#ifndef CFSRW_VIEW_H
#define CFSRW_VIEW_H

#include "refcnt.h"
#include "dhash.h"
#include "serialize.h"
#include "cfsrw_prot.h"

class view_constructor {
private:
  friend class view;
  chordID _root;
  vec<chordID> _head;
  vec<chordID> _latest;
  void sort ();
  void parse (xin *in);

public:
  view_constructor () {}
  view_constructor (const view &);

  void root (chordID n) { _root = n; }
  void add_log (chordID h, chordID b);
  str serialize ();
  chordID store (callback<void, dhash_stat, chordID>::ref cb);
};

class view {
private:
  chordID _root;
  vec<chordID> _head;
  vec<chordID> _latest;

  struct view_state {
    vec<chordID> latest;
    vec<unsigned> seqnum;
    unsigned nrecords;
    cfsrw_status error;
    view_state () { nrecords = 0; error = CFSRW_OK; }
  };

  void got_view (ptr<view_constructor> vc, 
                 callback<void, cfsrw_status>::ref cb, ptr<dhash_block> blk);
  void load_record (bool ishead, ptr<view_state> v, unsigned i,
                    callback<void, vervec, cfsrw_status>::ref cb, recordp e);
  void load_done (ptr<view_state> v,
                  callback<void, vervec, cfsrw_status>::ref cb);

public:
  view (callback<void, cfsrw_status>::ref cb, chordID addr);
  view (ptr<view_constructor> c);
  void take (ptr<view_constructor> c);

  unsigned nlogs () const { return _head.size(); }
  chordID root () const { return _root; }
  chordID head (unsigned i) const { return _head[i]; }
  chordID latest (unsigned i) const { return _latest[i]; }

  void version (callback<void, vervec, cfsrw_status>::ref cb);
};

#endif

