#include "snapshot.h"

void
init_attr (sfsro_inode_reg *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

void
init_attr (sfsro_inode_lnk *inode)
{
  inode->mtime.seconds = 0;
  inode->mtime.nseconds = 0;
  inode->attr.size = 0;
  inode->attr.mode = 0;
  inode->attr.uid = 0;
  inode->attr.gid = 0;
  inode->attr.rdev.major = 0;
  inode->attr.rdev.minor = 0;
}

void
set_attr (sfsro_inode_reg *inode, cfsrw_sattr *ca)
{
  if (ca->valid_mtime) {
    inode->mtime.seconds = ca->mtime;
    inode->mtime.nseconds = 0;
  }
  if (ca->valid_size)
    inode->attr.size = ca->size;
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
  if (ca->valid_uid)
    inode->attr.uid = ca->uid;
  if (ca->valid_gid)
    inode->attr.gid = ca->gid;
}

void
set_attr (sfsro_inode_lnk *inode, cfsrw_sattr *ca)
{
  if (ca->valid_mtime) {
    inode->mtime.seconds = ca->mtime;
    inode->mtime.nseconds = 0;
  }
  if (ca->valid_size)
    inode->attr.size = ca->size; // XXX what do i do?
  if (ca->valid_mode)
    inode->attr.mode = ca->mode;
  if (ca->valid_uid)
    inode->attr.uid = ca->uid;
  if (ca->valid_gid)
    inode->attr.gid = ca->gid;
}

void 
get_random_ID (chordID *ID, uint len) 
{
  char xx[len];
  rnd.getbytes (xx, len);
  mpz_set_rawmag_be (ID, xx, len);
}

str
get_dirname (nfspath3 path)
{
  unsigned short *a;
  a = (unsigned short*) path.cstr();
  unsigned short d = (*a) % 240;
  return strbuf ("/var/tmp/cfsrw-scache/%d", d);
}

str 
get_filename (nfspath3 path)
{
  unsigned short *a;
  a = (unsigned short*) path.cstr();
  unsigned short d = (*a) % 240;
  return strbuf ("/var/tmp/cfsrw-scache/%d/%s", d, path.cstr());
}

void 
set_cfsrwattr (cfsrw_sattr *sa, sfsro_inode_lnk *lnode)
{
  sa->valid_size = 1;
  sa->size = lnode->attr.size;
  sa->valid_mode = 1;
  sa->mode = lnode->attr.mode;
  sa->valid_uid = 1;
  sa->uid = lnode->attr.uid;
  sa->valid_gid = 1;
  sa->gid = lnode->attr.gid;
}

void 
set_cfsrwattr (cfsrw_sattr *sa, sfsro_inode_reg *rnode)
{
  sa->valid_size = 1;
  sa->size = rnode->attr.size;
  sa->valid_mode = 1;
  sa->mode = rnode->attr.mode;
  sa->valid_uid = 1;
  sa->uid = rnode->attr.uid;
  sa->valid_gid = 1;
  sa->gid = rnode->attr.gid;
}

void
combine_attr (ptr<cfsrw_states::attr> ca, ptr<sfsro_data> sdata)
{
  sfsro_inode *inode = sdata->inode;

  if (!ca->truncated) {
    if (inode->type == SFSROLNK && 
	(ca->sattr.size < inode->lnk->attr.size || !ca->sattr.valid_size))
      ca->sattr.size = inode->lnk->attr.size;
    else if (inode->type != SFSROLNK &&
	(ca->sattr.size < inode->reg->attr.size || !ca->sattr.valid_size))
      ca->sattr.size = inode->reg->attr.size;
  }
  if (inode->type == SFSROLNK) {
    sfsro_inode_lnk *lnode= inode->lnk;
    ca->nlink += lnode->nlink;
    ca->sattr.mode = ca->sattr.valid_mode ? ca->sattr.mode : lnode->attr.mode;
    ca->sattr.uid = ca->sattr.valid_uid ? ca->sattr.uid : lnode->attr.uid;
    ca->sattr.gid = ca->sattr.valid_gid ? ca->sattr.gid : lnode->attr.gid;
    if (!ca->sattr.valid_mtime || ca->sattr.mtime < lnode->mtime.seconds) {
      ca->sattr.mtime = lnode->mtime.seconds;
      ca->sattr.valid_mtime = true;
    }
    if (ca->ctime < lnode->ctime.seconds)
      ca->ctime = lnode->ctime.seconds;
  }
  else {
    sfsro_inode_reg *rnode= inode->reg;
    ca->nlink += rnode->nlink;
    ca->sattr.mode = ca->sattr.valid_mode ? ca->sattr.mode : rnode->attr.mode;
    ca->sattr.uid = ca->sattr.valid_uid ? ca->sattr.uid : rnode->attr.uid;
    ca->sattr.gid = ca->sattr.valid_gid ? ca->sattr.gid : rnode->attr.gid;
    if (!ca->sattr.valid_mtime || ca->sattr.mtime < rnode->mtime.seconds) {
      ca->sattr.mtime = rnode->mtime.seconds;
      ca->sattr.valid_mtime = true;
    }
    if (ca->ctime < rnode->ctime.seconds)
      ca->ctime = rnode->ctime.seconds;
  }
  ca->sattr.valid_size = 1;
  ca->sattr.valid_mode = 1;
  ca->sattr.valid_uid = 1;
  ca->sattr.valid_gid = 1;
  ca->sattr.valid_mtime = 1;
}

ptr<log_record>
set_log_record (bool is_lookup, ptr<sfsro_data> sdata, chordID dir, str name)
{  
  chordID inum;
  sfsro_inode *inode = sdata->inode;
  cfsrw_sattr sa;

  cfsrw_ftype itype;
  if (inode->type == SFSROLNK) {
    str dearm = dearmor64A ((*(*inode).lnk).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    itype = CFSRW_LNK;
    set_cfsrwattr (&sa, inode->lnk); // this is not used?
  } 
  else {
    str dearm = dearmor64A ((*(*inode).reg).path.cstr ());
    mpz_set_rawmag_be(&inum, dearm.cstr(), dearm.len());
    if (inode->type == SFSROREG || inode->type == SFSROREG_EXEC)
      itype = CFSRW_REG;
    else
      itype = CFSRW_DIR;
    set_cfsrwattr (&sa, inode->reg); // this is not used?
  }

  ptr<log_record> le;
  if (is_lookup) {
    itype = CFSRW_REG;
    if (inode->type == SFSROLNK)
      itype = CFSRW_LNK;
    else if (inode->type == SFSRODIR)
      itype = CFSRW_DIR;
    le = New refcounted <link_log_record> (inum, itype, dir, name);
  }
  else
    le = New refcounted <inode_log_record> (itype, inum, sa);

  return le;
}

bool 
dup_entry (rpc_ptr<entry3> *dep, sfsro_dirent *e)
{
  if (!dep) 
    return false;
  rpc_ptr<entry3> de = *dep;
  while (de && de->name != e->name)
    de = de->nextentry;
  if (de)
    return true;
  return false;
}

ref<dbrec>
id2dbrec(chordID id)
{
  str whipme = id.getraw ();
  void *key = (void *)whipme.cstr ();
  int len = whipme.len ();

  ptr<dbrec> q = New refcounted<dbrec> (key, len);
  return q;
}

chordID
dbrec2id (ptr<dbrec> r)
{
  str raw = str ( (char *)r->value, r->len);
  chordID ret;
  ret.setraw (raw);
  return ret;
}

int
dbcompare (ref<dbrec> a, ref<dbrec> b)
{
  chordID ax = dbrec2id (a);
  chordID bx = dbrec2id (b);
  if (ax < bx)
    return -1;
  else if (ax == bx)
    return 0;
  else
    return 1;
}

